/* -*- Mode: javascript; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
/* ***** BEGIN LICENSE BLOCK *****
 *	 Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is unityfox.
 *
 * The Initial Developer of the Original Code is
 * Canonical Ltd.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Chris Coulson <chris.coulson@canonical.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */

var EXPORTED_SYMBOLS = [ "LibUnity" ];

let Cu = Components.utils;
let Cc = Components.classes;
let Ci = Components.interfaces;

let nsIXULAppInfo = Ci.nsIXULAppInfo;

Cu.import("resource://gre/modules/ctypes.jsm");
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");

XPCOMUtils.defineLazyGetter(this, "libunity", function() {
  var libunity = ctypes.open("libunity.so.4");
  if (!libunity)
    throw "libunity is unavailable";

  return libunity;
});

XPCOMUtils.defineLazyGetter(this, "libgobject", function() {
  var libgobject = ctypes.open("libgobject-2.0.so.0");
  if (!libgobject)
    throw "libgobject is unavailable";

  return libgobject;
});

XPCOMUtils.defineLazyGetter(this, "UnityLauncherEntryPtr", function() {
  return ctypes.PointerType(ctypes.StructType("UnityLauncherEntry"));
});

XPCOMUtils.defineLazyGetter(this, "UnityInspectorPtr", function() {
  return ctypes.PointerType(ctypes.StructType("UnityInspector"));
});

XPCOMUtils.defineLazyGetter(this, "gboolean", function() {
  return ctypes.int;
});

XPCOMUtils.defineLazyGetter(this, "FALSE", function() {
  return 0;
});

XPCOMUtils.defineLazyGetter(this, "TRUE", function() {
  return 1;
});

XPCOMUtils.defineLazyGetter(this, "unity_launcher_entry_get_for_desktop_id", function() {
  var unity_launcher_entry_get_for_desktop_id =
    libunity.declare("unity_launcher_entry_get_for_desktop_id",
                     ctypes.default_abi,
                     UnityLauncherEntryPtr,
                     ctypes.char.ptr);
  if (!unity_launcher_entry_get_for_desktop_id)
    throw "unity_launcher_entry_get_for_desktop_id is unavailable";

  return unity_launcher_entry_get_for_desktop_id;
});

XPCOMUtils.defineLazyGetter(this, "unity_launcher_entry_set_count", function() {
  var unity_launcher_entry_set_count =
    libunity.declare("unity_launcher_entry_set_count",
                     ctypes.default_abi,
                     ctypes.void_t,
                     UnityLauncherEntryPtr,
                     ctypes.int64_t);
  if (!unity_launcher_entry_set_count)
    throw "unity_launcher_entry_set_count is unavailable";

  return unity_launcher_entry_set_count;
});

XPCOMUtils.defineLazyGetter(this, "unity_launcher_entry_set_count_visible", function() {
  var unity_launcher_entry_set_count_visible =
    libunity.declare("unity_launcher_entry_set_count_visible",
                     ctypes.default_abi,
                     ctypes.void_t,
                     UnityLauncherEntryPtr,
                     gboolean);
  if (!unity_launcher_entry_set_count_visible)
    return "unity_launcher_entry_set_count_visible is unavailable";

  return unity_launcher_entry_set_count_visible;
});

XPCOMUtils.defineLazyGetter(this, "unity_launcher_entry_set_progress", function() {
  var unity_launcher_entry_set_progress =
    libunity.declare("unity_launcher_entry_set_progress",
                     ctypes.default_abi,
                     ctypes.void_t,
                     UnityLauncherEntryPtr,
                     ctypes.double);
  if (!unity_launcher_entry_set_progress)
    throw "unity_launcher_entry_set_progress is unavailable";

  return unity_launcher_entry_set_progress;
});

XPCOMUtils.defineLazyGetter(this, "unity_launcher_entry_set_progress_visible", function() {
  var unity_launcher_entry_set_progress_visible =
    libunity.declare("unity_launcher_entry_set_progress_visible",
                     ctypes.default_abi,
                     ctypes.void_t,
                     UnityLauncherEntryPtr,
                     gboolean);
  if (!unity_launcher_entry_set_progress_visible)
    throw "unity_launcher_entry_set_progress_visible is unavailable";

  return unity_launcher_entry_set_progress_visible;
});

XPCOMUtils.defineLazyGetter(this, "unity_inspector_get_default", function() {
  var unity_inspector_get_default =
    libunity.declare("unity_inspector_get_default",
                     ctypes.default_abi,
                     UnityInspectorPtr);
  if (!unity_inspector_get_default)
    throw "unity_inspector_get_default is unavailable";

  return unity_inspector_get_default;
});

XPCOMUtils.defineLazyGetter(this, "unity_inspector_get_unity_running", function() {
  var unity_inspector_get_unity_running =
    libunity.declare("unity_inspector_get_unity_running",
                     ctypes.default_abi,
                     gboolean,
                     UnityInspectorPtr);
  if (!unity_inspector_get_unity_running)
    throw "unity_inspector_get_unity_running is unavailable";

  return unity_inspector_get_unity_running;
});

XPCOMUtils.defineLazyGetter(this, "g_object_unref", function() {
  var g_object_unref =
    libgobject.declare("g_object_unref",
                       ctypes.default_abi,
                       ctypes.void_t,
                       ctypes.void_t.ptr);
  if (!g_object_unref)
    throw "g_object_unref is unavailable";

  return g_object_unref;
});

var LibUnityPrivate = {

  get entry() {
    if (this._entry)
      return this._entry;

    var appName = Cc["@mozilla.org/xre/app-info;1"].
                    getService(nsIXULAppInfo).name.toLowerCase();
    this._entry =
      unity_launcher_entry_get_for_desktop_id(appName + ".desktop");
    if (!this._entry)
      throw "Failed to create UnityLauncherEntry";

    Services.obs.addObserver(this, "xpcom-will-shutdown", false);

    return this._entry;
  },

  set entry(aEntry) {
    if (this._entry)
      g_object_unref(this._entry);

    this._entry = aEntry;
  },

  observe: function LUP_observe(aSubject, aTopic, aData) {
    if (aTopic == "xpcom-will-shutdown") {
      this.entry = null;
    }
  },

  setCount: function LUP_setCount(aCount) {
    if (aCount === null) {
      unity_launcher_entry_set_count_visible(this.entry, FALSE);
    } else {
      unity_launcher_entry_set_count(this.entry, aCount);
      unity_launcher_entry_set_count_visible(this.entry, TRUE);
    }
  },

  setProgress: function LUP_setProgress(aProgress) {
    if (aProgress === null) {
      unity_launcher_entry_set_progress_visible(this.entry, FALSE);
    } else {
      unity_launcher_entry_set_progress(this.entry, aProgress);
      unity_launcher_entry_set_progress_visible(this.entry, TRUE);
    }
  },

  isAvailable: function LUP_isAvailable() {
    var ret = false;
    try {
      var inspector = unity_inspector_get_default();
      if (inspector) {
        ret = unity_inspector_get_unity_running(inspector) ==
          FALSE ? false : true;
        g_object_unref(inspector);
      }
    } catch(e) { } finally {
      return ret;
    }
  }
};

var LibUnity = {

  /**
    * Sets a count on this applications launcher icon
    *
    * @param aCount
    *        The integer value of the count to display, or null to
    *        hide the counter
    */
  setCount: function LU_setCount(aCount) {
    LibUnityPrivate.setCount(aCount);
  },

  /**
    * Sets a progress indicator on this applications launcher icon
    *
    * @param aProgress
    *        A float value between 0 and 1 representing the percentage
    *        progress, or null to hide the progress indicator
    */
  setProgress: function LU_setProgress(aProgress) {
    LibUnityPrivate.setProgress(aProgress);
  },

  /**
    * Checks if Unity is available and running
    *
    * @return true if Unity is available and running, else false
    *
    */
  isAvailable: function LU_isAvailable() {
    return LibUnityPrivate.isAvailable();
  }
};
