/* -*- Mode: javascript; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 *	 Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is messagingmenu-extension
 *
 * The Initial Developer of the Original Code is
 * Mozilla Messaging, Ltd.
 * Portions created by the Initial Developer are Copyright (C) 2010
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Mike Conley <mconley@mozillamessaging.com>
 *    Chris Coulson <chris.coulson@canonical.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */

var EXPORTED_SYMBOLS = ["MessagingMenu"];
var gWindow = null;

const Cc = Components.classes;
const Ci = Components.interfaces;
const Cu = Components.utils;

Cu.import("resource://gre/modules/ctypes.jsm");
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/NetUtil.jsm");
Cu.import("resource://gre/modules/FileUtils.jsm");
Cu.import("resource://gre/modules/AddonManager.jsm");
Cu.import("resource:///modules/mailServices.js");
Cu.import("resource:///modules/iteratorUtils.jsm");
Cu.import("resource://messagingmenu/LibGObject.jsm");
Cu.import("resource://messagingmenu/LibDbusmenu.jsm");
Cu.import("resource://messagingmenu/LibIndicate.jsm");
Cu.import("resource://messagingmenu/LibUnity.jsm");

// I need the GdkWindow lib to do focus hacking
Cu.import("resource://messagingmenu/LibGdkWindow.jsm");

["LOG", "WARN", "ERROR"].forEach(function(aName) {
  this.__defineGetter__(aName, function() {
    Components.utils.import("resource://gre/modules/AddonLogging.jsm");

    LogManager.getLogger("messagingmenu", this);
    return this[aName];
  });
}, this);

const FLDR_UNINTERESTING = Ci.nsMsgFolderFlags.Trash
                           | Ci.nsMsgFolderFlags.Junk
                           | Ci.nsMsgFolderFlags.SentMail
                           | Ci.nsMsgFolderFlags.Drafts
                           | Ci.nsMsgFolderFlags.Templates
                           | Ci.nsMsgFolderFlags.Queue
                           | Ci.nsMsgFolderFlags.Archive;

const FOLDER_URL_KEY                    = "url";
const SHELL_EXECUTABLES                 = ["thunderbird", "thunderbird-bin"];
const USER_SHARE_APPLICATIONS           = "/usr/share/applications/";
const SYSTEM_LAUNCHER_ENTRIES           = "/usr/share/indicators/messages/applications/";
const USER_LAUNCHER_ENTRIES             = ".config/indicators/messages/applications/";
const USER_BLACKLIST_ENTRIES            = ".config/indicators/messages/applications-blacklist/";
const MAX_INDICATORS                    = 6;
const ADDON_ID                          = "messagingmenu@mozilla.com";
const PREF_ROOT                         = "extensions.messagingmenu.";
const PREF_INCLUDE_NEWSGROUPS           = "includeNewsgroups";
const PREF_INCLUDE_RSS                  = "includeRSS";
const PREF_ENABLED                      = "enabled";
const PREF_USER_LAUNCHER_PATH           = "userLauncherPath";
const PREF_USER_LAUNCHER_MTIME          = "userLauncherMTime";
const PREF_ATTENTION_FOR_ALL            = "attentionForAll";
const NS_PREFBRANCH_PREFCHANGE_TOPIC_ID = "nsPref:changed";

var open3PaneCallback;
var contactsCallback;
var composerCallback;
var clickIndicatorCallback;

var injectTimestamp = function MM_injectTimestamp(aTimestamp) {
  let atts = new LibGdkWindow.GdkWindowAttributes;
  atts.window_type = 2;
  atts.x = atts.y = 0;
  atts.width = atts.height = 1;
  atts.wclass = 1;
  atts.event_mask = 0;
  let win = LibGdkWindow.GdkWindowNew(null, atts.address(), 0);
  LibGdkWindow.GdkX11WindowSetUserTime(win, aTimestamp);
  LibGdkWindow.GdkWindowDestroy(win);
}

var openAndFocus3Pane = function MM_openAndFocus3Pane(aInstance, aTimestamp, aUserData) {
  injectTimestamp(aTimestamp);
  MessagingMenu.tbProc.run(false, [], 0);
}

var openAndFocusAddressBook = function MM_openAndFocusAddressBook(aInstance, aTimestamp, aUserData) {
  LOG("Opening addressbook");
  injectTimestamp(aTimestamp);
  MessagingMenu.tbProc.run(false, ['-addressbook'], 1);
}

var openAndFocusComposer = function MM_openAndFocusComposer(aInstance, aTimestamp, aUserData) {
  LOG("Opening composer");
  injectTimestamp(aTimestamp);
  MessagingMenu.tbProc.run(false, ['-compose'], 1);
}

var onClickIndicator = function MM_onClickIndicator(aInstance, aTimestamp, aUserData) {
  let indicator = LibIndicate.IndicatorPtrCast(aInstance);
  let folderURL = LibIndicate.IndicatorGetProperty(indicator, FOLDER_URL_KEY).readString();
  LOG("Received click event on indicator for folder " + folderURL);

  let indicatorEntry = MessagingMenu.mIndicators[folderURL];
  if (!indicatorEntry) {
    WARN("No indicator for folder " + folderURL);
    return;
  }

  // Hide the indicator
  MessagingMenu.hideIndicator(indicatorEntry);

  var msg = MessagingMenu.messenger.msgHdrFromURI(indicatorEntry.messageURL);
  if(!msg) {
    WARN("Invalid message URI " + indicatorEntry.messageURL);
    return;
  }

  // Focus 3pane
  openAndFocus3Pane(aInstance, aTimestamp, aUserData);
  gWindow.document.getElementById("tabmail").switchToTab(0);
  gWindow.gFolderTreeView.selectFolder(msg.folder);
  gWindow.gFolderDisplay.selectMessage(msg);
}

function hasMultipleAccounts() {
  let count = 0;
  // We don't want to just call Count() on the account nsISupportsArray, as we
  // want to filter out accounts with "none" as the incoming server type
  // (eg, for Local Folders)
  for (let account in fixIterator(MailServices.accounts.accounts, Ci.nsIMsgAccount)) {
    if (account.incomingServer.type != "none") {
      count++
    }
  }

  return count > 1;
}

// Helper class to wrap a nsIPrefBranch and allow the caller
// to specify default values to be used where the pref doesn't exist
function Prefs(aBranch) {
  this.branch = aBranch;
}

Prefs.prototype = {
  getBoolPref: function P_getBoolPref(aName, aDefaultValue) {
    try {
      return this.branch.getBoolPref(aName);
    } catch(e) {
      return aDefaultValue;
    }
  },

  setCharPref: function P_setCharPref(aName, aValue) {
    this.branch.setCharPref(aName, aValue);
  },

  getCharPref: function P_getCharPref(aName, aDefaultValue) {
    try {
      return this.branch.getCharPref(aName);
    } catch(e) {
      return aDefaultValue;
    }
  },

  setIntPrefAsChar: function P_setIntPrefAsChar(aName, aValue) {
    this.setCharPref(aName, aValue.toString());
  },

  getIntPrefFromChar: function P_getIntPrefFromChar(aName, aDefaultValue) {
    return parseInt(this.getCharPref(aName, aDefaultValue.toString()));
  },

  clearUserPref: function P_clearUserPref(aName) {
    this.branch.clearUserPref(aName);
  },

  addObserver: function P_addObserver(aDomain, aObserver, aHoldWeak) {
    this.branch.QueryInterface(Ci.nsIPrefBranch2)
      .addObserver(aDomain, aObserver, aHoldWeak);
  },

  removeObserver: function P_removeObserver(aDomain, aObserver) {
    this.branch.QueryInterface(Ci.nsIPrefBranch2)
      .removeObserver(aDomain, aObserver);
  }
};

// Small helper class which takes a directory containing messaging menu
// launcher entries and tells the listener whether one of them is ours
function LauncherEntryFinder(aDir, aDesktopFile, aListener, aMethod) {
  LOG("Searching for launcher entry for " + aDesktopFile + " in " + aDir.path);
  if (!aDir.exists() || !aDir.isDirectory()) {
    WARN(aDir.path + " does not exist or is not a directory");
    aListener[aMethod](aDir, null);
    return;
  }

  this.listener = aListener;
  this.desktopFile = aDesktopFile;
  this.entries = aDir.directoryEntries;
  this.method = aMethod;
  this.dir = aDir;

  this.processNextEntry();
}

LauncherEntryFinder.prototype = {
  processNextEntry: function MMEF_processNextEntry() {
    if (this.entries.hasMoreElements()) {
      var entry = this.entries.getNext().QueryInterface(Ci.nsIFile);
      if (!entry.isFile()) {
        this.processNextEntry();
      }
      var self = this;
      NetUtil.asyncFetch(entry, function(inputStream, status) {
        let data = NetUtil.readInputStreamToString(inputStream, inputStream.available());
        if (data.replace(/\n$/,"") == self.desktopFile) {
          LOG("Found launcher entry " + entry.path);
          self.listener[self.method](self.dir, entry);
        } else {
          self.processNextEntry();
        }
      });
    } else {
      LOG("No launcher entry found");
      this.listener[this.method](this.dir, null);
    }
  }
};

/* Given a particular message header, determines whether or not
 * it's something that's worth showing in the Messaging Menu.
 *
 * @param aItemHeader An nsIMsgDBHdr for a message.
 * @param aCallback A function to call if the message should 
 *                  be indicated to the user
 */
function MMIndicatorMessageCheck (aItemHeader, aCallback) {
  // FIXME:  This function needs a bit of cleanup - I think the plinko-style
  // boolean flag stuff could be done a bit better.
  // See bug 806123:  https://bugs.launchpad.net/messagingmenu-extension/+bug/806123
  let folder = aItemHeader.folder;
  let shouldIndicate = false;
  let shouldRequestAtt = false;
  let forceRequestAtt = false;

  LOG("Checking if message " + aItemHeader.folder.getUriForMsg(aItemHeader) +
      " in " + folder.folderURL + " is worth indicating");

  // Filter out the new messages that don't count...like junk.
  var junkScore = aItemHeader.getStringProperty("junkscore");
  if ((junkScore != "") && (junkScore != "0")) {
    // We're junk.  Not worth indicating.
    LOG("Rejecting message with junkscore = " + junkScore);
    return;
  }

  if (folder.flags & FLDR_UNINTERESTING) {
    LOG("Rejecting message with flags = " + folder.flags.toString());
    return;
  }

  if (aItemHeader.isRead) {
    // The item has already been read
    LOG("Rejecting message which has already been read");
    return;
  }

  if (folder.flags & Ci.nsMsgFolderFlags.Mail) {
    LOG("Accepting indication for mail message");
    shouldIndicate = true;
  }

  // Are we checking news groups?
  if (!shouldIndicate &&
    MessagingMenu.prefs.getBoolPref(PREF_INCLUDE_NEWSGROUPS, true)) {
      if (folder.flags & Ci.nsMsgFolderFlags.Newsgroup) {
        LOG("Accepting indication for newsgroup message");
        shouldIndicate = true;
    }
  }

  if (!shouldIndicate &&
      MessagingMenu.prefs.getBoolPref(PREF_INCLUDE_RSS, true)) {
    if (folder.server.type == "rss") {
      LOG("Accepting indication for message from RSS feed");
      shouldIndicate = true;
    }
  }

  if (!shouldIndicate) {
    LOG("Message with flags " + folder.flags.toString() + " rejected");
    return;
  }

  // At this point, we know shouldIndicate is true.

  if (MessagingMenu.prefs.getBoolPref(PREF_ATTENTION_FOR_ALL,
                                      false)) {
    LOG("Requesting attention for all indications");
    shouldRequestAtt = true;
    forceRequestAtt = true;
  }

  if (!shouldRequestAtt && aItemHeader.isFlagged) {
    LOG("Requesting attention for flagged message");
    shouldRequestAtt = true;
  }

  if (!shouldRequestAtt) {
    let recipients = aItemHeader.recipients.split(",");
    let re = /.*<([^>]*)>/; // Convert "Foo <bar>" in to "bar"
    for (let i in recipients) {
      let recipient = recipients[i].replace(re, "$1");
      if (shouldRequestAtt) {
        break;
      }

      for (let id in fixIterator(MailServices.accounts.allIdentities,
                                 Ci.nsIMsgIdentity)) {
        if (recipient.indexOf(id.email) != -1) {
          LOG("Requesting attention for message with recipient " + recipient);
          shouldRequestAtt = true;
          break;
        }
      }
    }
  }

  if (!shouldRequestAtt &&
      (aItemHeader.priority >= Ci.nsMsgPriority.high)) {
    LOG("Requesting attention for message with priority = " +
         aItemHeader.priority.toString());
    shouldRequestAtt = true;
  }

  if (shouldRequestAtt && !forceRequestAtt &&
      aItemHeader.priority <= Ci.nsMsgPriority.low &&
      aItemHeader.priority > Ci.nsMsgPriority.none) {
    LOG("Cancelling request for attention on message with priority = " +
        aItemHeader.priority.toString());
    shouldRequestAtt = false;
  }

  aCallback(shouldRequestAtt);
}

function MMIndicatorEntry (aFolder) {
  LOG("Creating indicator for folder " + aFolder.folderURL);
  this.folder = aFolder;
  this.indicator = LibIndicate.IndicatorNew();
  if (hasMultipleAccounts()) {
    this.label = aFolder.prettiestName +
                 " (" + aFolder.server.prettyName + ")";
  } else {
    this.label = aFolder.prettiestName;
  }
  this.newCount = 0;
  this.dateInSeconds = 0;
  this.cancelAttention();
  this.hide();

  LibGObject.g_signal_connect(this.indicator, "user-display",
                              clickIndicatorCallback, null);

  LibIndicate.IndicatorSetProperty(this.indicator,
                                   FOLDER_URL_KEY,
                                   aFolder.folderURL);

  Services.prefs.addObserver("mail.accountmanager.accounts",
                             this, false);
}

MMIndicatorEntry.prototype = {
  get indicator() {
    if (this._indicator) {
      return this._indicator;
    }

    throw "No IndicateIndicator. Have we been destroyed?";
  },

  set indicator(aIndicator) {
    this._indicator = aIndicator;
  },

  getAttention: function MMIE_getAttention() {
    if (!this.active) {
      WARN("Attempting to request attention for an inactive indicator");
      return;
    }

    LOG("Requesting attention for folder " + this.folder.folderURL);
    LibIndicate.IndicatorSetProperty(this.indicator,
                                     LibIndicate.INDICATOR_MESSAGES_PROP_ATTENTION,
                                     "true");
    this._attention = true;
  },

  cancelAttention: function MMIE_cancelAttention() {
    LOG("Cancelling attention for folder " + this.folder.folderURL);
    LibIndicate.IndicatorSetProperty(this.indicator,
                                     LibIndicate.INDICATOR_MESSAGES_PROP_ATTENTION,
                                     "false");
    this._attention = false;
  },

  hasAttention: function MMIE_hasAttention() {
    return this._attention;
  },

  set label(aName) {
    LOG("Setting label for folder " + this.folder.folderURL + " to " + aName);
    LibIndicate.IndicatorSetProperty(this.indicator,
                                     LibIndicate.INDICATOR_MESSAGES_PROP_NAME,
                                     aName);
  },

  set newCount(aCount) {
    LOG("Setting unread count for folder " + this.folder.folderURL +
        " to " + aCount.toString());
    LibIndicate.IndicatorSetProperty(this.indicator,
                                     LibIndicate.INDICATOR_MESSAGES_PROP_COUNT,
                                     aCount.toString());
    this._newCount = aCount;
  },

  get newCount() {
    return this._newCount;
  },

  set active(aActive) {
    LOG("Setting indicator for folder " + this.folder.folderURL +
        (aActive ? " active" : " inactive"));
    this._active = aActive;
  },

  get active() {
    return this._active;
  },

  show: function MMIE_show() {
    if (!this.active) {
      WARN("Attempting to display an inactive indicator");
      return;
    }

    LOG("Showing indicator for folder " + this.folder.folderURL);
    LibIndicate.IndicatorShow(this.indicator);
  },

  hide: function MMIE_hide() {
    LOG("Hiding indicator for folder " + this.folder.folderURL);
    LibIndicate.IndicatorHide(this.indicator);
  },

  get visible() {
    return LibIndicate.IndicatorIsVisible(this.indicator) != 0;
  },

  get priority() {
    let score = 0;

    if (this.folder.flags & Ci.nsMsgFolderFlags.Inbox) {
      score += 3;
    }

    if (this.hasAttention()) {
      score += 1;
    }

    return score;
  },

  // We consider an indicator to be less important if:
  // 1) It has a lower priority, or
  // 2) It has the same priority and is more recent
  hasPriorityOver: function MMIE_hasPriorityOver(aIndicator) {
    return ((aIndicator.priority < this.priority) ||
            ((aIndicator.dateInSeconds > this.dateInSeconds) &&
             (aIndicator.priority == this.priority))) ? true : false;
  },

  destroy: function MMIE_destroy() {
    LOG("Destroying indicator for folder " + this.folder.folderURL);
    LibGObject.g_object_unref(this.indicator);
    this.indicator = null;

    Services.prefs.removeObserver("mail.accountmanager.accounts",
                                  this);
  },

  observe: function MMIE_observe(subject, topic, data) {
    // An account was added or removed. Note that this observer fires
    // before nsIMsgAccountManager is up-to-date, so we add the next
    // bit to the event loop
    LOG("Account settings updated. Updating label for folder " +
        this.folder.folderURL);
    var self = this;
    gWindow.setTimeout(function() {
      if (hasMultipleAccounts()) {
        self.label = self.folder.prettiestName +
                     " (" + self.folder.server.prettyName + ")";
      } else {
        self.label = self.folder.prettiestName;
      }
    }, 0);
  }
};

var MessagingMenu = {
  initialized: false,
  enabled: false,
  mIndicators: {},
  visibleIndicators: 0,
  _badgeCount: 0,

  get messenger() {
    if (this._messenger)
      return this._messenger;
    this._messenger = Cc["@mozilla.org/messenger;1"].createInstance()
                      .QueryInterface(Ci.nsIMessenger);
    return this._messenger;
  },

  get desktopFile() {
    if (this._desktopFile)
      return this._desktopFile;

    var appName = Services.appinfo.name.toLowerCase();
    this._desktopFile = USER_SHARE_APPLICATIONS + appName + ".desktop";
    return this._desktopFile;
  },

  get tbProc() {
    if (this._tbProc)
      return this._tbProc;
      
    this._tbProc = Cc["@mozilla.org/process/util;1"].createInstance(Ci.nsIProcess);
    var tbExec = Services.dirsvc.get("CurProcD", Ci.nsILocalFile);
    try {
      for each (let executable in SHELL_EXECUTABLES) {
        let tbExecTry = tbExec.clone();
        tbExecTry.append(executable);
        if (tbExecTry.exists() && tbExecTry.isExecutable()) {
          tbExec.append(executable);
          break;
        }
      }
      if (!tbExec.exists() || !tbExec.isExecutable()) {
        throw("Could not find a Thunderbird executable at " + tbExec.path);
      }
      this._tbProc.init(tbExec);
    } catch(e) {
      Cu.reportError("Could not init Messaging Menu support: " + e);
    }
    return this._tbProc;
  },

  get prefs() {
    if (this._prefs)
      return this._prefs;
    this._prefs = new Prefs(Services.prefs.getBranch(PREF_ROOT));
    return this._prefs;
  },

  get indicateServer() {
    if (this._indicateServer)
      return this._indicateServer;

    let indicateServer = LibIndicate.ServerRefDefault();
    LibIndicate.ServerSetType(indicateServer, "message.email");
    LibIndicate.ServerSetDesktopFile(indicateServer, 
                                     this.desktopFile);

    open3PaneCallback = LibGObject.GCallbackFunction(openAndFocus3Pane);

    LibGObject.g_signal_connect(indicateServer, "server-display",
        open3PaneCallback, null);

    let bundle = Services.strings.createBundle(
                  "chrome://messagingmenu/locale/messagingmenu.properties");

    let server = LibDbusmenu.ServerNew("/messaging/commands");
    let root = LibDbusmenu.MenuitemNew();

    let composeMi = LibDbusmenu.MenuitemNew();
    LibDbusmenu.MenuitemPropertySet(composeMi, "label",
                                    bundle.GetStringFromName("composeNewMessage"));
    LibDbusmenu.MenuitemPropertySetBool(composeMi, "visible", true);
    composerCallback = LibGObject.GCallbackFunction(openAndFocusComposer);

    LibGObject.g_signal_connect(composeMi,
                                LibDbusmenu.MENUITEM_SIGNAL_ITEM_ACTIVATED, 
                                composerCallback, null);
    LibDbusmenu.MenuitemChildAppend(root, composeMi);

    let contactsMi = LibDbusmenu.MenuitemNew();
    LibDbusmenu.MenuitemPropertySet(contactsMi, "label",
                                    bundle.GetStringFromName("contacts"));
    LibDbusmenu.MenuitemPropertySetBool(contactsMi, "visible", true);
    contactsCallback = LibGObject.GCallbackFunction(openAndFocusAddressBook);

    LibGObject.g_signal_connect(contactsMi,
                                LibDbusmenu.MENUITEM_SIGNAL_ITEM_ACTIVATED, 
                                contactsCallback, null);
    LibDbusmenu.MenuitemChildAppend(root, contactsMi);

    LibDbusmenu.ServerSetRoot(server, root);
    LibIndicate.ServerSetMenu(indicateServer, server); 
    LibIndicate.ServerShow(indicateServer);

    clickIndicatorCallback = LibGObject.GCallbackFunction(onClickIndicator);

    this._indicateServer = indicateServer;
    return this._indicateServer;
  },

  get badgeCount() {
    return this._badgeCount;
  },

  set badgeCount(aCount) {
    LOG("Setting total new count to " + aCount.toString());
    this._badgeCount = aCount;
    if (aCount > 0) {
      LibUnity.setCount(aCount);
    } else {
      LibUnity.setCount(null);
    }
  },

  sysLauncherFindResult: function MM_sysLauncherFindResult(aDir, aFound) {
    if (!aFound) {
      // There is no system-provided static launcher entry for us in the
      // messaging menu

      let userLauncherEntriesDir = Services.dirsvc.get("Home", Ci.nsILocalFile);
      userLauncherEntriesDir.appendRelativePath(USER_LAUNCHER_ENTRIES);
      new LauncherEntryFinder(userLauncherEntriesDir, this.desktopFile, this,
                              "createLauncherEntryIfNoneExists");
    } 
  },

  createLauncherEntryIfNoneExists: function MM_createLauncherEntryIfNoneExists(
                                            aDir, aFound) {
    if (!aFound) {
      let entry = aDir;
      entry.append(Services.appinfo.name.toLowerCase());
      let ostream = FileUtils.openSafeFileOutputStream(entry,
                                                       FileUtils.MODE_WRONLY |
                                                       FileUtils.MODE_CREATE |
                                                       FileUtils.MODE_TRUNCATE);
      let converter = Cc["@mozilla.org/intl/scriptableunicodeconverter"]
                      .createInstance(Ci.nsIScriptableUnicodeConverter);
      converter.charset = "UTF-8";
      let istream = converter.convertToInputStream(this.desktopFile);
      var self = this;
      NetUtil.asyncCopy(istream, ostream, function() {
        self.prefs.setCharPref(PREF_USER_LAUNCHER_PATH, entry.path);
        self.prefs.setIntPrefAsChar(PREF_USER_LAUNCHER_MTIME,
                                    entry.lastModifiedTime);
      });
    }
  },

  init: function MM_init(aWindow) {
    if (this.initialized)
      return;

    LOG("Initializing MessagingMenu");

    gWindow = aWindow;

    // Check if we have a static launcher entry in the messaging menu. If we
    // don't, then we should add one and also display "Contacts" and "Compose"
    // menu items. If there is one, then it was probably added by the Thunderbird
    // package. We don't need to create one or show the extra menu items in that case
    let sysLauncherEntriesDir = Cc["@mozilla.org/file/local;1"]
                                .createInstance(Ci.nsILocalFile);
    sysLauncherEntriesDir.initWithPath(SYSTEM_LAUNCHER_ENTRIES);
    new LauncherEntryFinder(sysLauncherEntriesDir, this.desktopFile, this,
                            "sysLauncherFindResult");

    AddonManager.addAddonListener(this);
    this.prefs.addObserver("", this, false);
    Services.obs.addObserver(this, "xpcom-will-shutdown", false);

    this.badgeCount = 0;

    if (this.prefs.getBoolPref(PREF_ENABLED, true)) {
      this.enable();
    } else {
      this.disableAndHide();
    }

    this.initialized = true;
  },

  removeLauncherEntry: function MM_removeLauncherEntry(aDir, aFound) {
    if (aFound) {
      LOG("Removing launcher entry " + aFound.path);
      aFound.remove(false);
    }
  },

  enable: function MM_enable() {
    if (this.enabled) {
      WARN("Trying to enable more than once");
      return;
    }

    LOG("Enabling messaging indicator");

    if (!this.indicateServer) {
      Cu.reportError("Could not construct the Messaging Menu server.");
      return;
    }

    let userBlacklistDir = Services.dirsvc.get("Home", Ci.nsILocalFile);
    userBlacklistDir.appendRelativePath(USER_BLACKLIST_ENTRIES);
    new LauncherEntryFinder(userBlacklistDir, this.desktopFile, this,
                            "removeLauncherEntry");

    let notificationFlags = Ci.nsIFolderListener.added
                            | Ci.nsIFolderListener.propertyFlagChanged;
    MailServices.mailSession.AddFolderListener(this, notificationFlags);

    this.enabled = true;
  },

  createBlacklistEntryIfNoneExists: function MM_createBlacklistEntryIfNoneExists(
                                             aDir, aFound) {
    if (aFound)
      return;

    let entry = aDir;
    entry.append(Services.appinfo.name.toLowerCase());
    let ostream = FileUtils.openSafeFileOutputStream(entry,
                                                     FileUtils.MODE_WRONLY |
                                                     FileUtils.MODE_CREATE |
                                                     FileUtils.MODE_TRUNCATE);
    let converter = Cc["@mozilla.org/intl/scriptableunicodeconverter"]
                    .createInstance(Ci.nsIScriptableUnicodeConverter);
    converter.charset = "UTF-8";
    let istream = converter.convertToInputStream(this.desktopFile);
    var self = this;
    NetUtil.asyncCopy(istream, ostream, null);
  },

  disableAndHide: function MM_disableAndHide() {
    LOG("Hiding messaging indicator");
    let userBlacklistDir = Services.dirsvc.get("Home", Ci.nsILocalFile);
    userBlacklistDir.appendRelativePath(USER_BLACKLIST_ENTRIES);
    new LauncherEntryFinder(userBlacklistDir, this.desktopFile, this,
                            "createBlacklistEntryIfNoneExists");

    this.disable();
  },

  cleanup: function MM_cleanup() {
    // We're being uninstalled or disabled. If we created a launcher
    // entry in the messaging menu, make sure we clean it up
    let userLauncherEntry = this.prefs.getCharPref(PREF_USER_LAUNCHER_PATH,
                                                   null);
    let userLauncherEntryMTime =
      this.prefs.getIntPrefFromChar(PREF_USER_LAUNCHER_MTIME, 0);
    this.prefs.clearUserPref(PREF_USER_LAUNCHER_PATH);
    this.prefs.clearUserPref(PREF_USER_LAUNCHER_MTIME);
    if (userLauncherEntry) {
      let userLauncherEntryFile = Cc["@mozilla.org/file/local;1"]
                                  .createInstance(Ci.nsILocalFile);
      userLauncherEntryFile.initWithPath(userLauncherEntry);
      if (userLauncherEntryFile.exists() &&
          userLauncherEntryFile.isFile() &&
          userLauncherEntryFile.lastModifiedTime == userLauncherEntryMTime) {
        LOG("Removing launcher entry at " + userLauncherEntry);
        userLauncherEntryFile.remove(false);
      }
    }
  },

  disable: function MM_disable() {
    if (!this.enabled)
      return;

    LOG("Disabling messaging indicator");

    MailServices.mailSession.RemoveFolderListener(this);

    // Remove references for any leftover indicators
    for (let key in this.mIndicators)
      this.mIndicators[key].destroy();
    this.mIndicators = {};
    this.visibleIndicators = 0;

    if (this._indicateServer)
      LibGObject.g_object_unref(this._indicateServer);
    this._indicateServer = null;

    this.badgeCount = 0;

    this.enabled = false;
  },

  shutdown: function MM_shutdown() {
    if (!this.initialized) {
      WARN("Calling shutdown before we are initialized");
      return;
    }

    LOG("Shutting down");

    this.disable();

    AddonManager.removeAddonListener(this);
    Services.prefs.removeObserver(this, "xpcom-will-shutdown");
    this.prefs.removeObserver("", this);

    this.initialized = false;
  },

  refreshBadgeCount: function MM_refreshBadgeCount() {
    let accumulator = 0;
    for (let url in this.mIndicators) {
      accumulator += this.mIndicators[url].newCount;
    }

    this.badgeCount = accumulator;
  },

  /* Given a message header, displays an indicator for it's folder
   * and requests attention
   *
   * @param aItemHeader A nsIMsgDBHdr for the message that we're
   *        trying to show in the Messaging Menu.
   */
  doIndication: function MM_doIndication(aItemHeader, aShouldRequestAtt) {
    let itemFolder = aItemHeader.folder;
    let folderURL = itemFolder.folderURL;
    LOG("Doing indication for folder " + folderURL);
    if (!this.mIndicators[folderURL]) {
      // Create an indicator for this folder if one doesn't already exist
      this.mIndicators[folderURL] = new MMIndicatorEntry(itemFolder);
    }

    let indicator = this.mIndicators[folderURL];
    indicator.newCount += 1;

    LOG("Current indicator dateInSeconds = " + indicator.dateInSeconds.toString());
    LOG("Message item dateInSeconds = " + aItemHeader.dateInSeconds.toString());
    if (indicator.active) {
      LOG("Indicator for folder is already active");
    }

    if (!indicator.active ||
        (indicator.dateInSeconds > aItemHeader.dateInSeconds)) {
      indicator.messageURL = itemFolder.getUriForMsg(aItemHeader);
      indicator.dateInSeconds = aItemHeader.dateInSeconds;
      indicator.active = true;
    }

    if (aShouldRequestAtt) {
      indicator.getAttention();
    }

    this.badgeCount += 1;

    if (indicator.visible) {
      LOG("Indicator is already visible");
      // No more to do
      return;
    }

    LOG("Indicator priority is " + indicator.priority.toString());

    // Don't show more than MAX_INDICATORS indicators
    if (this.visibleIndicators < MAX_INDICATORS) {
      this.visibleIndicators++;
      indicator.show();
      LOG("There are now " + this.visibleIndicators.toString() +
          " visible indicators");
    } else {
      // We are already displaying MAX_INDICATORS. Lets see if one of the
      // current ones can be bumped off, to make way for the new one
      LOG("There are already " + MAX_INDICATORS.toString() + " visible indicators");
      let doomedIndicator = null

      // This will make your head explode, but basically, what we want to do
      // is iterate over the currently displayed indicator entries and see if
      // one of them should make way for the new indicator.
      for (let url in this.mIndicators) {
        let existingIndicator = this.mIndicators[url];
        // This one already isn't visible, so don't care...
        if (!existingIndicator.visible) {
          continue;
        }

        let refIndicator = doomedIndicator ? doomedIndicator : indicator;

        if (refIndicator.hasPriorityOver(existingIndicator)) {
          LOG("Indicator with priority=" + existingIndicator.priority.toString() +
              " and dateInSeconds=" + existingIndicator.dateInSeconds + " is " +
              " a candidate for hiding");
          doomedIndicator = existingIndicator;
        }
      }

      if (doomedIndicator) {
        doomedIndicator.hide();
        indicator.show();
      }
    }
  },

  stopIndication: function MM_stopIndication(aItemHeader) {
    let folderURL = aItemHeader.folder.folderURL;
    LOG("Stopping indication for folder " + folderURL);
     
    if (!this.mIndicators[folderURL])
      return;

    this.hideIndicator(this.mIndicators[folderURL]);
  },

  hideIndicator: function MM_hideIndicator(aIndicator) {
    if (aIndicator.visible) {
      aIndicator.hide();
      if (this.visibleIndicators > 0) {
        this.visibleIndicators--;
        LOG("There are now " + this.visibleIndicators.toString() +
            " visible indicators");
      } else {
        ERROR("Invalid MessagingMenu.visibleIndicators count: "
              + this.visibleIndicators.toString());
      }
    }

    let savedCount = aIndicator.newCount;
    aIndicator.cancelAttention();
    aIndicator.newCount = 0;
    aIndicator.active = false;

    let oldBadgeCount = this.badgeCount;
    this.refreshBadgeCount();
    if (oldBadgeCount != (this.badgeCount + savedCount)) {
      WARN("The badge count got out of sync with the actual number of new messages");
    }

    if (this.visibleIndicators >= MAX_INDICATORS) {
      return;
    }

    // Now see if there are any pending indicators to be shown.
    // If there are more than one, we give priority to the indicator
    // with the highest priority. If there are more than one of those,
    // then we give priority to the one which has been waiting the longest
    let pendingIndicator = null;
    for (let url in this.mIndicators) {
      let indicator = this.mIndicators[url];
      if (indicator.active && !indicator.visible &&
          (!pendingIndicator ||
           indicator.hasPriorityOver(pendingIndicator))) {

        pendingIndicator = indicator;
      }
    }

    if (pendingIndicator) {
      pendingIndicator.show();
      this.visibleIndicators++;
      LOG("There are now " + this.visibleIndicators.toString() +
          " visible indicators");
    }
  },

  /* Observes when items are added to folders, and when
   * message flags change.  Also listens for notifications
   * sent by uMessagingMenuService for opening messages based
   * on an Indicator that was clicked.
   */
  OnItemAdded: function MM_OnItemAdded(parentItem, item) {
    let header = item.QueryInterface(Ci.nsIMsgDBHdr);
    let self = this;
    MMIndicatorMessageCheck(header, function(aShouldRequestAtt) {
      self.doIndication(header, aShouldRequestAtt);
    });
  },

  OnItemPropertyFlagChanged: function MM_OnItemPropertyFlagChanged(item,
                                                                   property,
                                                                   oldFlag,
                                                                   newFlag) {
    if((oldFlag & Ci.nsMsgMessageFlags.New)
        && !(newFlag & Ci.nsMsgMessageFlags.New)) {
      let header = item.QueryInterface(Ci.nsIMsgDBHdr);
      this.stopIndication(header);
    }
  },

  observe: function(aSubject, aTopic, aData) {
    if (aTopic == "xpcom-will-shutdown") {
      LOG("Got shutdown notification");
      this.shutdown();
    } else if (aTopic == NS_PREFBRANCH_PREFCHANGE_TOPIC_ID) {
      LOG("Got prefchange notification for " + aData);
      if (aData == PREF_ENABLED) {
        let prefs = new Prefs(aSubject.QueryInterface(Ci.nsIPrefBranch));
        let enabled = prefs.getBoolPref(aData, true);
        if (enabled) {
          this.enable();
        } else {
          this.disableAndHide();
        }
      }
    } else {
      WARN("Observer notification not intended for us: " + aTopic);
    }
  },

  onUninstalling: function(aAddon, aNeedsRestart) {
    if (aAddon.id == ADDON_ID) {
      LOG("Addon is being uninstalled");
      this.cleanup();
    }
  },

  onDisabling: function(aAddon, aNeedsRestart) {
    if (aAddon.id == ADDON_ID) {
      LOG("Addon is being disabled");
      this.disableAndHide();
    }
  },

  onEnabling: function(aAddon, aNeedsRestart) {
    if (aAddon.id == ADDON_ID) {
      LOG("Addon is being enabled");
      this.enable();
    }
  }
}

